from zeroconf import ServiceBrowser, ServiceInfo, ServiceListener, Zeroconf


class MyListener(ServiceListener):

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        print(f"Service {name} updated")

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        print(f"Service {name} removed")

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        info = zc.get_service_info(type_, name)
        print(f"Service {name} added, service info: {info}")


zeroconf = Zeroconf()
listener = MyListener()
browser = ServiceBrowser(zeroconf, "_smartenergy._tcp.local.", listener)

si = ServiceInfo(type_="_smartenergy._tcp.local.",
                 name="device-dev1._smartenergy._tcp.local.",
                 properties={
                     'txtvers': 1,
                     'https': 7443,
                     'dcap': '/dcap'
                 })
si.addresses = ['127.0.0.1']
si.port = 7443
zeroconf.register_service(si)
try:
    input("Press enter to exit...\n\n")
finally:
    zeroconf.close()


from typing import Awaitable, Callable, Dict, Union

from nicegui import background_tasks, ui
from nicegui.dependencies import register_component

register_component('router_frame', __file__, 'router_frame.js')


class Router():

    def __init__(self) -> None:
        self.routes: Dict[str, Callable] = {}
        self.content: ui.element = None

    def add(self, path: str):
        def decorator(func: Callable):
            self.routes[path] = func
            return func
        return decorator
    
    def add_route(self, path: str, func: Callable):
        self.routes[path] = func

    def open(self, target: Union[Callable, str]):
        if isinstance(target, str):
            path = target
            builder = self.routes[target]
        else:
            path = {v: k for k, v in self.routes.items()}[target]
            builder = target

        async def build():
            with self.content:
                await ui.run_javascript(f'history.pushState({{page: "{path}"}}, "", "{path}")', respond=False)
                result = builder()
                if isinstance(result, Awaitable):
                    await result
        self.content.clear()
        background_tasks.create(build())

    def frame(self) -> ui.element:
        self.content = ui.element('router_frame').on('open', lambda msg: self.open(msg['args']))
        return self.content
    
global_router = Router()


import dataclasses
import json
import logging
from typing import List

from nicegui import ui

import ieee_2030_5.models as m
from ieee_2030_5.models.enums import CurveType, DERUnitRefType

from ..pages import Pages, show_global_header
from ..session import get_curve_list, send_curve

_log = logging.getLogger(__name__)

curve_data_points: List[m.CurveData] = [m.CurveData() for x in range(10)]
curve_list: m.DERCurveList = m.DERCurveList()
current_curve: m.DERCurve = m.DERCurve()
der_ref_type: DERUnitRefType = DERUnitRefType.NA
updating_curve: bool = False

@ui.refreshable
def render_select():
    with ui.row():
        der_curves = [c.description for c in curve_list.DERCurve]
        der_curves.insert(0, "NEW")
        if not current_curve.description:
            ui.select(der_curves, label="Curves", value=der_curves[0],
                      on_change=lambda e: change_der_curve(e.value)).classes("w-64")
        else:
            ui.select(der_curves, label="Curves", value=current_curve.description,
                      on_change=lambda e: change_der_curve(e.value)).classes("w-64")
        if current_curve.href:
            ui.label(f"Href: {current_curve.href}")

def change_der_curve_type(new_curve_type: DERUnitRefType):
    pass

def change_der_curve(new_curve_description: str):
    global curve_data_points, current_curve, der_ref_type, updating_curve
    
    updating_curve = True
    # Create a new item.
    if new_curve_description is None or new_curve_description == "NEW":
        current_curve = m.DERCurve()
        curve_data_points = []
        der_ref_type = DERUnitRefType.NA
    else:
        for curve in curve_list.DERCurve:
            if curve.description == new_curve_description:
                current_curve = curve
                curve_data_points = curve.CurveData
                break
            
    # We know this is a brand new curve so we create the max number of curve points
    # allowed so the user can enter data.
    if len(current_curve.CurveData) == 0:
        curve_data_points = [m.CurveData() for x in range(10)]
    else:
        # We know there is some data already so we want to keep that and then
        # add the maximum number of items availalbe.
        curve_data_points = [dataclasses.replace(x) for x in current_curve.CurveData]
        while len(curve_data_points) < 10:
            curve_data_points.append(m.CurveData())
    
    assert len(curve_data_points) == 10    
    render_curve_form.refresh()
    render_curve_data.refresh()
    updating_curve = False
    # render_curve_form()    
    # render_curve_data()

@ui.refreshable  
def render_curve_data():
    with ui.row():
        for i in range(10):
            with ui.column():
                ui.checkbox(f"excitation-{i}").bind_value(curve_data_points[i], "excitation")
                ui.number(f"x-value-{i}").props("size=10").bind_value(curve_data_points[i], "xvalue")
                ui.number(f"y-value-{i}").props("size=10").bind_value(curve_data_points[i], "yvalue")
                
def curve_type_refresh(select: ui.select):
    _log.debug(f"select: {select.value}")
    if current_curve.curveType != select.value:
        current_curve.curveType = select.value
        render_curve_form.refresh()
        
@ui.refreshable
def render_curve_form():
    _log.debug("Rendering curve form")
    with ui.row():
        with ui.column():            
            curveType = ui.select({curve_type.value: curve_type.name for index, curve_type in enumerate(CurveType)},
                                  label="Curve Type").classes("w-64").bind_value_from(current_curve, "curveType")
            curveType.on_value_change = lambda e: curve_type_refresh(curveType)
    with ui.row():
        with ui.column():
            mRID = ui.input('mRID').bind_value(current_curve, "mRID")
        with ui.column():
            description = ui.input('description').bind_value(current_curve, 'description')
        with ui.column():
            version = ui.input('version').bind_value(current_curve, "version")
            
    if current_curve.curveType == CurveType.opModVoltVar:
        with ui.row():
            if current_curve.autonomousVRefEnable is None:
                current_curve.autonomousVRefEnable = False
                
            with ui.column():
                autonomousVRefEnable = ui.checkbox("Enable", value=False).bind_value(current_curve, "autonomousVRefEnable")       
            with ui.column():
                autonomousVRefTimeConstant = ui.number("autonomousVRefTimeConstant").bind_value(current_curve, "autonomousVRefTimeConstant").classes("w-64") \
                                                                                    .bind_enabled(autonomousVRefEnable, "value")
                
    with ui.row():
        with ui.column():
            openLoopTms = ui.number("openLoopTms").bind_value(current_curve, "openLoopTms")
        with ui.column():
            rampDecTms = ui.number("rampDecTms").bind_value(current_curve, "rampDecTms")
        with ui.column():
            rampIncTms = ui.number("rampIncTms").bind_value(current_curve, "rampIncTms")
        with ui.column():
            rampPT1Tms = ui.number("rampPT1Tms").bind_value(current_curve, "rampPT1Tms")
        with ui.column():
            vRef = ui.input("vRef").bind_value(current_curve, "vRef")
    with ui.row():
        with ui.column():
            yRefType = ui.select({ v.value: v.name for index, v in enumerate(DERUnitRefType)},
                                 label="DER Unit Ref Type").bind_value(current_curve, "yRefType").classes("w-64")
        with ui.column():
            xMultiplier = ui.number("xMultiplier").bind_value(current_curve, "xMultiplier").classes("w-24")
        with ui.column():
            yMultiplier = ui.number("yMultiplier").bind_value(current_curve, "yMultiplier").classes("w-24")
        

    
data_row = None  

def render_object_data():
    global data_row, curve_data_points
    if data_row is None:
        data_row = ui.row()
    else:
        data_row.clear()
    with data_row:
        with ui.column():
            ui.label("Curve Data")
            for curve_data in curve_data_points:
                ui.markdown(f"""{dataclasses.asdict(curve_data)}""")
        with ui.column():
            ui.label("DER Curve")
            ui.markdown(f"""
```json
{json.dumps(dataclasses.asdict(current_curve), indent=4)}
```
""")
                
def revert_changes():
    pass

def validate_and_submit():
    """Validate the objects and submit them to the server.
    
    Returns:
        bool: True if successful, False otherwise.
    """
    global curve_data_points, current_curve
    current_curve.CurveData = curve_data_points
    
    if append_new_curve := current_curve.href is None:
        ...
    
    # POST or PUT the curve to the server     
    dc = send_curve(current_curve)
    
    assert isinstance(dc, m.DERCurve)
    
    # update the variables for state
    curve_data_points = dc.CurveData
    current_curve = dc
    
    # IF its post then we append to the curve list
    if append_new_curve:
        curve_list.DERCurve.append(dc)
    
    render_select.refresh()
    
    ui.notify("Curve saved successfully")
    
    

@ui.page("/curves")
def show_curves():
    global curve_list
    curve_response = get_curve_list()
    if curve_response:
        curve_list = curve_response
    
    show_global_header(Pages.CURVES)
    
    render_select()
    ui.separator()
    render_curve_form()
    ui.separator()
    render_curve_data()
    
    
    with ui.row():
        ui.button("Store", on_click=validate_and_submit)
        ui.button("Revert", on_click=revert_changes)
        # Create a "newline style" separator
        ui.separator().clear()
    

    
    
    # excitation: boolean [0..1]
    # xvalue: Int32
    # yvalue: Int32


import flet as ft

from typing import Callable

import logging

_log = logging.getLogger(__name__)


def home_view(page: ft.Page) -> ft.View:
    return ft.View(
        "/",
        [
            ft.AppBar(title=ft.Text("2030.5 GUI"),
                      automatically_imply_leading=False,
                      bgcolor=ft.colors.SURFACE_VARIANT),
            ft.ElevatedButton("Server Control", on_click=lambda _: page.go("/server_control")),
            ft.ElevatedButton("Show List Tree", on_click=lambda _: page.go("/show_tree")),
    #ElevatedButton("Subscriber App", on_click=lambda _: page.go("/subscriber")),
        ])



import base64
import uuid
from dataclasses import dataclass, fields
from pathlib import Path
from typing import Optional, Type

from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.xml import XmlParser
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata.formats.dataclass.parsers.handlers import LxmlEventHandler

from ieee_2030_5.models.sep import EndDevice, EndDeviceList

__xml_context__ = XmlContext()
__parser_config__ = ParserConfig(fail_on_unknown_attributes=True, fail_on_unknown_properties=True)
__xml_parser__ = XmlParser(config=__parser_config__,
                           context=__xml_context__,
                           handler=LxmlEventHandler)
__config__ = SerializerConfig(xml_declaration=False, pretty_print=True)
__serializer__ = XmlSerializer(config=__config__)
__ns_map__ = {None: "urn:ieee:std:2030.5:ns"}

import ieee_2030_5.types_ as t
import ieee_2030_5.utils as tls


class PrivateKeyDeosntExist(Exception):

    def __init__(self, private_key_path: Path):
        super().__init__()
        self.pk_path = private_key_path

    def __str__(self) -> str:
        return f"The path {self.pk_path} does not exist!"


class CertExistsError(Exception):

    def __init__(self, cert_path: Path):
        super().__init__()
        self.cert_path = cert_path

    def __str__(self) -> str:
        return f"The path {self.cert_path} already exists!"


class CADoesNotExist(Exception):

    def __str__(self) -> str:
        return "The CA certificate does not exist!"


def serialize_dataclass(obj: dataclass) -> str:
    """
    Serializes a dataclass that was created via xsdata to an xml string for
    returning to a client.
    """
    return __serializer__.render(obj, ns_map=__ns_map__)


def xml_to_dataclass(xml: str, type: Optional[Type] = None) -> dataclass:
    """
    Parse the xml passed and return result from loaded classes.
    """
    parsed = __xml_parser__.from_string(xml, type)

    # The xml parser from string seems to double decode the lfDI which
    # probably means I am doing something wrong.  However, this fixes
    # the issue and it is correct after we encode the lFDI.  I will
    # do the same with other entities as needed.
    if isinstance(parsed, EndDevice) and parsed.lFDI:
        parsed.lFDI = base64.b16encode(parsed.lFDI)
    elif isinstance(parsed, EndDeviceList):
        for ed in parsed.EndDevice:
            if ed.lFDI:
                ed.lFDI = base64.b16encode(ed.lFDI)

    return parsed


def dataclass_to_xml(dc: dataclass) -> str:
    return serialize_dataclass(dc)


def get_lfdi_from_cert(path: Path) -> t.Lfdi:
    """
    Using the fingerprint of the certifcate return the left truncation of 160 bits with no check digit.
    Example:
      From:
        3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213A
      Return:
        3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5
        as an integer.
    """

    # 160 / 4 == 40
    fp = tls.OpensslWrapper.tls_get_fingerprint_from_cert(path)
    fp = fp.replace(":", "")
    lfdi = t.Lfdi(fp[:40])
    return lfdi


def get_sfdi_from_lfdi(lfdi: t.Lfdi) -> int:
    """

    Args:
        lfdi:

    Returns:

    """
    from ieee_2030_5.certs import sfdi_from_lfdi
    return sfdi_from_lfdi(lfdi)


def uuid_2030_5() -> str:
    return str(uuid.uuid4()).replace('-', '').upper()


class TLSWrap:

    @staticmethod
    def tls_create_private_key(file_path: Path):
        """
        Creates a private key in the path that is specified.  The path will be overwritten
        if it already exists.

        Args:
            file_path:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_ca_certificate(common_name: str, private_key_file: Path, ca_cert_file: Path):
        """
        Create a ca certificate from using common name private key and ca certificate file.

        Args:
            common_name:
            private_key_file:
            ca_cert_file:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_csr(common_name: str, private_key_file: Path, server_csr_file: Path):
        """

        Args:
            common_name:
            private_key_file:
            server_csr_file:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_signed_certificate(common_name: str,
                                      ca_key_file: Path,
                                      ca_cert_file: Path,
                                      private_key_file: Path,
                                      cert_file: Path,
                                      as_server: bool = False):
        """

        Args:
            common_name:
            ca_key_file:
            ca_cert_file:
            private_key_file:
            cert_file:
            as_server:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_get_fingerprint_from_cert(cert_file: Path, algorithm: str = "sha256") -> str:
        """

        Args:
            cert_file:
            algorithm:

        Returns:

        """

    @staticmethod
    def tls_create_pkcs23_pem_and_cert(private_key_file: Path, cert_file: Path,
                                       combined_file: Path):
        """

        Args:
            private_key_file:
            cert_file:
            combined_file:

        Returns:

        """
        raise NotImplementedError()


from ieee_2030_5.utils.tls_wrapper import OpensslWrapper
from ieee_2030_5.utils.cryptography_wrapper import CryptographyWrapper

__all__ = ['OpensslWrapper', 'CryptographyWrapper']



from __future__ import annotations
import os

import inspect
from pprint import pprint
import logging
import typing
from copy import deepcopy
from dataclasses import dataclass, fields, is_dataclass, asdict
from enum import Enum
from pathlib import Path
from typing import (Any, ClassVar, Dict, Generic, List, Optional, Protocol, Type, TypeVar, Union,
                    get_args, get_origin)

import yaml
from blinker import Signal

import ieee_2030_5.config as cfg
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository

_log = logging.getLogger(__name__)


class AlreadyExists(Exception):
    pass


class NotFoundError(Exception):
    pass


load_event = Signal("load-store-event")
store_event = Signal("store-data-event")


def __get_store__(store_name: str) -> Path:
    if cfg.ServerConfiguration.storage_path is None:
        cfg.ServerConfiguration.storage_path = Path("data_store")
    elif isinstance(cfg.ServerConfiguration.storage_path, str):
        cfg.ServerConfiguration.storage_path = Path(cfg.ServerConfiguration.storage_path)

    store_path = cfg.ServerConfiguration.storage_path
    store_path.mkdir(parents=True, exist_ok=True)
    store_path = store_path / f"{store_name}.yml"

    return store_path


def do_load_event(caller: Union[Adapter, ResourceListAdapter]) -> None:
    """Load an adaptor type from the data_store path.


    """
    store_file = None
    if isinstance(caller, Adapter):
        _log.debug(f"Loading store {caller.generic_type_name}")
        store_file = __get_store__(caller.generic_type_name)
    elif isinstance(caller, ResourceListAdapter):
        store_file = __get_store__(caller.__class__.__name__)
    else:
        raise ValueError(f"Invalid caller type {type(caller)}")

    if not store_file.exists():
        _log.debug(f"Store {store_file.as_posix()} does not exist at present.")
        return

    # Load from yaml unsafe values etc.
    with open(store_file, "r") as f:
        items = yaml.load(f, Loader=yaml.UnsafeLoader)

        caller.__dict__.update(items)

    _log.debug(f"Loaded {caller.count} items from store")


def do_save_event(caller: Union[Adapter, ResourceListAdapter]) -> None:

    store_file = None
    if isinstance(caller, Adapter):
        _log.debug(f"Loading store {caller.generic_type_name}")
        store_file = __get_store__(caller.generic_type_name)
        _log.debug(f"Storing: {caller.generic_type_name}")
    elif isinstance(caller, ResourceListAdapter):
        store_file = __get_store__(caller.__class__.__name__)
    else:
        raise ValueError(f"Invalid caller type {type(caller)}")

    with open(store_file, 'w') as f:
        yaml.dump(caller.__dict__, f, default_flow_style=False, allow_unicode=True)


load_event.connect(do_load_event)
store_event.connect(do_save_event)


class ReturnCode(Enum):
    OK = 200
    CREATED = 201
    NO_CONTENT = 204
    BAD_REQUEST = 400


def populate_from_kwargs(obj: object, **kwargs) -> Dict[str, Any]:

    if not is_dataclass(obj):
        raise ValueError(f"The passed object {obj} is not a dataclass.")

    for k in fields(obj):
        if k.name in kwargs:
            type_eval = eval(k.type)

            if typing.get_args(type_eval) is typing.get_args(Optional[int]):
                setattr(obj, k.name, int(kwargs[k.name]))
            elif typing.get_args(k.type) is typing.get_args(Optional[bool]):
                setattr(obj, k.name, bool(kwargs[k.name]))
            # elif bytes in args:
            #     setattr(obj, k.name, bytes(kwargs[k.name]))
            else:
                setattr(obj, k.name, kwargs[k.name])
            kwargs.pop(k.name)
    return kwargs


class AdapterIndexProtocol(Protocol):

    def fetch_at(self, index: int) -> m.Resource:
        pass


class AdapterListProtocol(AdapterIndexProtocol):

    def fetch_list(self, start: int = 0, after: int = 0, limit: int = 0) -> m.List_type:
        pass

    def fetch_edev_all(self) -> List:
        pass


ready_signal = Signal("ready-signal")

T = TypeVar('T')
C = TypeVar('C')
D = TypeVar('D')


class ResourceListAdapter:
    """
    A generic list adapter class for storing and retrieving lists of objects of a specific type. The adapter is
    initialized with an empty list of URLs and an empty dictionary of list containers. The adapter provides methods for
    adding URLs to the list, retrieving lists of objects from the adapter, and registering types for use with the
    adapter. The adapter also provides a `load_event` signal that is emitted when the adapter is loaded.

    :ivar _list_urls: A list of URLs for the adapter
    :vartype _list_urls: list
    :ivar _list_containers: A dictionary of list containers, indexed by URL
    :vartype _list_containers: dict
    :ivar _types: A dictionary of types registered with the adapter, indexed by type name
    :vartype _types: dict
    """

    def __init__(self):
        self._list_urls = []
        self._container_dict: Dict[str, Dict[int, D]] = {}
        self._singleton_dict: Dict[str, D] = {}
        self._types: Dict[str, D] = {}
        if not os.environ.get('IEEE_ADAPTER_IGNORE_INITIAL_LOAD'):
            _log.debug(f"Intializing adapter {self.__class__.__name__}")
            load_event.send(self)
        else:
            _log.debug(
                f"Skip loading initial store due to IEEE_ADAPTER_IGNORE_INITIAL_LOAD being set")

    def list_size(self, list_uri: str) -> int:
        alist = self._container_dict.get(list_uri, [])
        return len(alist)

    def count(self) -> int:
        count_of = 0
        for v in self._container_dict.values():
            count_of += len(v)
        return count_of

    def get_type(self, list_uri: str) -> D:
        return self._types.get(list_uri)

    def initialize_uri(self, list_uri: str, obj: D):
        if self._container_dict.get(list_uri) and self._types.get(list_uri) != obj:
            _log.error("Must initialize before container has any items.")
            raise ValueError("Must initialize before container has any items.")
        self._types[list_uri] = obj

    def append_and_increment_href(self, list_uri: str, obj: D) -> D:
        url_parts: list[str] = list_uri.split(hrefs.SEP)
        try:
            int(url_parts[0])
            obj.href = hrefs.SEP.join(url_parts[1:].append(str(self.list_size(list_uri))))
        except ValueError:
            obj.href = hrefs.SEP.join([list_uri, str(self.list_size(list_uri))])

        self.append(list_uri, obj)
        return obj

    def append(self, list_uri: str, obj: D):
        """
        Appends an object to a list container in the adapter.

        The following rules are used to determine the container of the passed url.

        1. If the list container does not exist, it is created.
        2. If the list container exists but has not been initialized with a type, the type of the object[D]
           is used to initialize the list container.
        3. If the list container exists and has been initialized with a type, the type of the object[D] is
           validated against the initialized type.  An ValueError is thrown if the types do not match.

        :param list_uri: The URI of the list container
        :type list_uri: str
        :param obj: The object to append to the list container
        :type obj: Generic[D] or DList container
        :raises AssertionError: If the type of the object does not match the initialized type of the list container
        """
        cls = obj.__class__
        if issubclass(cls, (m.List_type, m.SubscribableList)):
            expected_type = eval(f'm.{cls.__name__[:cls.__name__.find("List")]}')

            if self._types.get(list_uri) is not None:
                raise ValueError(f"List for {list_uri} has already been initialized")

            self._types[list_uri] = expected_type

            # Recurse over the list appending to the end for each in the list
            for ele in getattr(obj, expected_type.__name__):
                self.append(list_uri, ele)

            # Exit here as all of the sub-items have been added now.
            return

        else:    # if there is a type
            expected_type = self._types.get(list_uri)

        if expected_type:
            if not isinstance(obj, expected_type):
                raise ValueError(f"Object {obj} is not of type {expected_type.__name__}")
        else:
            self.initialize_uri(list_uri, obj.__class__)

        if list_uri not in self._container_dict:
            self._container_dict[list_uri] = {}
        self._container_dict[list_uri][len(self._container_dict[list_uri])] = obj
        store_event.send(self)

    def get_by_mrid(self, list_uri: str, mrid: str) -> Optional[T]:
        try:
            return self.get_item_by_prop(list_uri, "mRID", mrid)
        except NotFoundError as ex:
            _log.warning(f"mRID: {mrid} not found in list: {list_uri}")
            return None

    def get_item_by_prop(self, list_uri: str, prop: str, value: Any) -> D:
        for item in self._container_dict.get(list_uri, {}).values():
            if getattr(item, prop) == value:
                return item
        raise NotFoundError(f"Uri {list_uri} does not contain {prop} == {value}")

    def has_list(self, list_uri: str) -> bool:
        return list_uri in self._container_dict

    def get_resource_list(self,
                          list_uri: str,
                          start: int = 0,
                          after: int = 0,
                          limit: int = 0,
                          sort_by: List[str] = [],
                          reverse: bool = False) -> Union[m.List_type, m.SubscribableList]:
        if isinstance(sort_by, str):
            sort_by = [sort_by]
        cls = self.get_type(list_uri)
        if cls is None:
            raise KeyError(f"Resource list {list_uri} not found in adapter")

        thelist = eval(f"m.{cls.__name__}List()")
        try:
            # Create a new list because the type is known and the uri is already known
            # there should be no need to add any items to the list, but the list should
            # exist on first access now.
            if list_uri not in self._container_dict:
                self._container_dict[list_uri] = {}
            thecontainerlist = list(self._container_dict[list_uri].values())
            for sort in sort_by:
                subobj = sort.split('.')
                if len(subobj) == 2:
                    try:
                        thecontainerlist = sorted(
                            thecontainerlist,
                            key=lambda o: getattr(getattr(o, subobj[0]), subobj[1]),
                            reverse=reverse)
                    except AttributeError:
                        # happens when value is none
                        pass
                elif len(subobj) == 1:
                    thecontainerlist = sorted(thecontainerlist,
                                              key=lambda o: getattr(o, sort),
                                              reverse=reverse)
                else:
                    raise ValueError("Can only sort through single nested properties.")
                # if "." in sort:

                # thecontainerlist = sorted(thecontainerlist, key=sort)
            thelist.href = list_uri
            thelist.all = len(thecontainerlist)
            if start == after == limit == 0:
                setattr(thelist, cls.__name__, thecontainerlist)
            else:
                posx = start + after
                setattr(thelist, cls.__name__, thecontainerlist[posx:posx + limit])
            thelist.results = len(getattr(thelist, cls.__name__))
        except KeyError:
            thelist.all = 0
            thelist.href = list_uri
            thelist.results = len(getattr(thelist, cls.__name__))
        return thelist

    def get_list(self, list_uri: str, start: int = 0, limit: int = 0, after: int = 0) -> D:
        if list_uri not in self._container_dict:
            raise KeyError(f"List {list_uri} not found in adapter")

        return list(self._container_dict[list_uri].values())

    def set_single(self, uri: str, obj: D):
        self._singleton_dict[uri] = obj

    def get_single(self, uri: str) -> D:
        return self._singleton_dict.get(uri)

    def get(self, list_uri: str, key: int) -> D:
        if list_uri not in self._container_dict:
            raise KeyError(f"List {list_uri} not found in adapter")
        if not isinstance(key, int):
            key = int(key)

        try:
            return self._container_dict[list_uri][key]
        except KeyError:
            raise NotFoundError(f"Key {key} not found in list {list_uri}")

    def set(self, list_uri: str, key: int, value: D, overwrite: bool = True) -> D:
        if list_uri not in self._container_dict:
            raise KeyError(f"List {list_uri} not found in adapter")

        if key in self._container_dict[list_uri] and not overwrite:
            raise AlreadyExists(
                f"Key {key} already exists in list {list_uri} but overwrite not set to True")

        self._container_dict[list_uri][key] = value
        store_event.send(self)

    def store(self):
        store_event.send(self)

    def get_values(self, list_uri: str, sort_by: Optional[str] = None) -> List[D]:
        raise ValueError("Hmmmmm refactoring.")
        cpy = deepcopy(list(self._container_dict[list_uri].values()))
        if sort_by is not None:
            return sorted(cpy, key=lambda x: getattr(x, sort_by))
        else:
            return cpy

    def remove(self, list_uri: str, index: int):
        del self._container_dict[list_uri][index]
        store_event.send(self)

    def render_container(self, list_uri: str, instance: object, prop: str):
        setattr(instance, prop, deepcopy(self._container_dict[list_uri]))

    def print_container(self, list_uri: str):
        pprint(self._container_dict[list_uri])

    def print_all(self):

        for k in sorted(self._container_dict.keys()):
            print("K:", k)
            for index, v in self._container_dict[k].items():
                print("index:", index)
                pprint(v.__dict__)
            #pprint(self._container_dict[k].)

    def get_all_as_dict(self) -> Dict[str, Dict[int, D]]:
        out = {}

        def replace_bytes(obj: dict) -> dict:
            for k, v in obj.items():
                if isinstance(v, dict):
                    obj[k] = replace_bytes(v)
                elif isinstance(v, bytearray):
                    obj[k] = v.hex()
                elif isinstance(v, bytes):
                    obj[k] = v.hex()
            return obj

        for k in sorted(self._container_dict.keys()):
            for index, v in self._container_dict[k].items():
                if k not in out:
                    out[k] = {}
                a = deepcopy(v)
                out[k] = asdict(a)
                out[k] = replace_bytes(out[k])

        data = dict(lists=out)

        single = {}

        for k, v in self._singleton_dict.items():
            single[k] = asdict(deepcopy(v))
            single[k] = replace_bytes(single[k])

        data['singletons'] = single

        return data

    def clear_all(self):
        self._container_dict.clear()
        self._list_urls.clear()
        self._types.clear()

    def clear(self, list_uri: str):
        if list_uri in self._container_dict:
            self._container_dict[list_uri].clear()
            self._list_urls[list_uri].clear()
            self._types[list_uri].clear()


class Adapter(Generic[T]):
    """
    A generic adapter class for storing and retrieving objects of a specific type. The adapter is initialized with a
    URL prefix and a generic type parameter. The adapter maintains an internal dictionary of objects, indexed by an
    integer ID. The adapter provides methods for adding, updating, and deleting objects, as well as retrieving objects
    by ID or by a custom filter function. The adapter also provides a `count` property that returns the number of
    objects currently stored in the adapter.

    :param url_prefix: The URL prefix for the adapter
    :type url_prefix: str
    :param kwargs: Additional keyword arguments
    :type kwargs: dict
    :raises ValueError: If the `generic_type` parameter is missing from `kwargs`
    """

    def __init__(self, url_prefix: str, **kwargs):
        if "generic_type" not in kwargs:
            raise ValueError("Missing generic_type parameter")
        self._generic_type: Type = kwargs['generic_type']
        self._href_prefix: str = url_prefix
        self._current_index: int = -1
        self._item_list: Dict[int, T] = {}
        if not os.environ.get('IEEE_ADAPTER_IGNORE_INITIAL_LOAD'):
            _log.debug(f"Intializing adapter {self.generic_type_name}")
            load_event.send(self)
        else:
            _log.debug(
                f"Skip loading initial store due to IEEE_ADAPTER_IGNORE_INITIAL_LOAD being set")

    @property
    def count(self) -> int:
        return len(self._item_list)

    @property
    def generic_type_name(self) -> str:
        return self._generic_type.__name__

    @property
    def href_prefix(self) -> str:
        return self._href_prefix

    @property
    def href(self) -> str:
        return self._href_prefix

    @href.setter
    def href(self, value: str) -> None:
        self._href_prefix = value

    def clear(self) -> None:
        self._current_index = -1
        self._item_list: Dict[int, T] = {}
        store_event.send(self)

    def fetch_by_mrid(self, mrid: str) -> Optional[T]:
        return self.fetch_by_property("mRID", mrid)

    def fetch_by_href(self, href: str) -> Optional[T]:
        return self.fetch_by_property("href", href)

    def fetch_by_property(self, prop: str, prop_value: Any) -> Optional[T]:
        for obj in self._item_list.values():
            # Most properties are pointers to other objects so we are going to
            # check both the property and the sub object property here, because
            # that should save some of the time later when we are looking for
            # hrefs and and can't get to them because they are wrapped in a
            # Link object.
            under_test = getattr(obj, prop)
            if isinstance(under_test, str):
                if under_test == prop_value:
                    return obj
            else:
                if getattr(under_test, prop) == prop_value:
                    return obj

    def add(self, item: T) -> T:
        if not isinstance(item, self._generic_type):
            raise ValueError(f"Item {item} is not of type {self._generic_type}")

        # Only replace if href is specified.
        if hasattr(item, 'href') and getattr(item, 'href') is None:
            setattr(item, 'href', hrefs.SEP.join([self._href_prefix,
                                                  str(self._current_index + 1)]))
        self._current_index += 1
        self._item_list[self._current_index] = item

        store_event.send(self)
        return item

    def fetch_all(self,
                  container: Optional[D] = None,
                  start: int = 0,
                  after: int = 0,
                  limit: int = 1) -> D:

        if container is not None:
            if not container.__class__.__name__.endswith("List"):
                raise ValueError("Must have List as the last portion of the name for instance")

            prop_found = container.__class__.__name__[:container.__class__.__name__.find("List")]

            items = list(self._item_list.values())
            all_len = len(items)
            all_results = len(items)
            all_items = items

            if start > len(items):
                all_items = []
                all_results = 0
            else:
                if limit == 0:
                    all_items = items[start:]
                else:
                    all_items = items[start:start + limit]
                all_results = len(all_items)

            setattr(container, prop_found, all_items)
            setattr(container, "all", all_len)
            setattr(container, "results", all_results)
        else:
            container = list(self._item_list.values())

        return container

    def fetch_index(self, obj: T, using_prop: str = None) -> int:
        found_index = -1
        for index, obj1 in self._item_list.items():
            if using_prop is None:
                if obj1 == obj:
                    found_index = index
                    break
            else:
                if getattr(obj, using_prop) == getattr(obj1, using_prop):
                    found_index = index
                    break
        if found_index == -1:
            raise KeyError(f"Object {obj} not found in adapter")
        return found_index

    def fetch(self, index: int):
        return self._item_list[index]

    def put(self, index: int, obj: T):
        self._item_list[index] = obj
        store_event.send(self)

    def fetch_by_mrid(self, mRID: str):
        for item in self._item_list.values():
            if not hasattr(item, 'mRID'):
                raise ValueError(f"Item of {type(T)} does not have mRID property")
            if item.mRID == mRID:
                return item

        raise KeyError(f"mRID ({mRID}) not found.")

    def size(self) -> int:
        return len(self._item_list)

    def __len__(self) -> int:
        return len(self._item_list)

    def store(self):
        store_event.send(self)

    def get_all_as_dict(self) -> Dict[str, Dict[int, D]]:
        out = {}

        def replace_bytes(obj: dict) -> dict:
            for k, v in obj.items():
                if isinstance(v, dict):
                    obj[k] = replace_bytes(v)
                elif isinstance(v, bytearray):
                    obj[k] = v.hex()
                elif isinstance(v, bytes):
                    obj[k] = v.hex()
            return obj

        for k in sorted(self._item_list.keys()):
            copyofitem = deepcopy(self._item_list[k])
            out[k] = asdict(copyofitem)
            out[k] = replace_bytes(out[k])

        return out



from ieee_2030_5.adapters.adapters import (DERAdapter, DERControlAdapter, DERCurveAdapter,
                                           DERProgramAdapter, DeviceCapabilityAdapter,
                                           EndDeviceAdapter, FunctionSetAssignmentsAdapter,
                                           RegistrationAdapter, TimeAdapter, ListAdapter,
                                           create_mirror_usage_point, create_or_update_meter_reading)

__all__ = [
    'DERControlAdapter', 'DERCurveAdapter', 'DERProgramAdapter', 'DeviceCapabilityAdapter',
    'EndDeviceAdapter', 'FunctionSetAssignmentsAdapter', 'RegistrationAdapter', 'DERAdapter',
    'TimeAdapter', 'create_mirror_usage_point', 'create_or_update_meter_reading', 'ListAdapter'
]


def clear_all_adapters():
    for adpt in __all__:
        obj = eval(adpt)
        if isinstance(obj, Adapter):
            obj.clear()
        elif isinstance(obj, ResourceListAdapter):
            obj.clear_all()


# from ieee_2030_5.adapters.log import LogAdapter
# from ieee_2030_5.adapters.mupupt import MirrorUsagePointAdapter



import flet as ft

from typing import Callable

import logging
import os
from pathlib import Path
from fastapi.middleware.wsgi import WSGIMiddleware

_log = logging.getLogger(__name__)

start_button: ft.Ref[ft.ElevatedButton] = ft.Ref()
stop_button: ft.Ref[ft.ElevatedButton] = ft.Ref()
this_view: ft.Ref[ft.View] = ft.Ref()
this_page: ft.Ref[ft.Page] = ft.Ref()



def start_server() -> None:
    _log.debug("Starting server")
    from ieee_2030_5_gui.__main__ import app
    from ieee_2030_5.flask_server import make_app

    config_file = Path(os.environ["CONFIG_FILE"])
    reset_certs = bool(os.environ.get("RESET_CERTS", False))
    flask_app = make_app(config_file=config_file, reset_certs=reset_certs)

    app.mount("/api",
              WSGIMiddleware(flask_app))

    start_button.current.disabled = True
    stop_button.current.disabled = False
    this_view.current.update()


    # app.run(host=os.getenv(f"{PREFIX}_HOST"), port=os.getenv(f"{PREFIX}_PORT"), debug=True, ssl_context=(os.getenv(f"{PREFIX}_CLIENT_CERT"), os.getenv(f"{PREFIX}_CLIENT_KEY"), os.getenv(f"{PREFIX}_CA_CERT")))
    # app.run(host=os.getenv(f"{PREFIX}_HOST"), port=os.getenv(f"{PREFIX}_PORT"), debug=True, ssl_context=(os.getenv(f"{PREFIX}_CLIENT_CERT"), os.getenv(f"{PREFIX}_CLIENT_KEY"), os.getenv(f"{PREFIX}_CA_CERT"))

def stop_server() -> None:
    _log.debug("Stopping server")
    app.unmount("/api")
    start_button.current.disabled = False
    stop_button.current.disabled = True
    this_view.current.update()

def server_control_view(page: ft.Page) -> ft.View:
    this_page.current = page

    this_view.current = ft.View("/", [
        ft.AppBar(title=ft.Text("Server Control"),
               automatically_imply_leading=False,
               bgcolor=ft.colors.SURFACE_VARIANT),
        ft.ElevatedButton("Start Server", ref=start_button, on_click=lambda _: start_server()),
        ft.ElevatedButton("Stop Server", disabled=True, ref=stop_button, on_click=lambda _: stop_server()),
    ])

    return this_view.current



# -------------------------------------------------------------------------------
# Copyright (c) 2022, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

import logging
import os
import shutil

log_level = os.environ.get('LOGGING_LEVEL', 'DEBUG').upper()

levels = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR
}

logging.basicConfig(level=levels[log_level])

import socket
import sys
import threading
from argparse import ArgumentParser
from multiprocessing import Process
from pathlib import Path
from time import sleep

import yaml
from werkzeug.serving import BaseWSGIServer

import ieee_2030_5.hrefs as hrefs
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import InvalidConfigFile, ServerConfiguration
from ieee_2030_5.data.indexer import add_href
from ieee_2030_5.server.server_constructs import initialize_2030_5
from ieee_2030_5.adapters.gridappsd_adapter import GridAPPSDAdapter

_log = logging.getLogger()


class ServerThread(threading.Thread):

    def __init__(self, server: BaseWSGIServer):
        threading.Thread.__init__(self, daemon=True)
        self.server = server

    def run(self):
        _log.info(f'starting server on {self.server.host}:{self.server.port}')
        self.server.serve_forever()

    def shutdown(self):
        _log.info("shutting down server")
        self.server.shutdown()


def get_tls_repository(cfg: ServerConfiguration,
                       create_certificates_for_devices: bool = True) -> TLSRepository:
    tlsrepo = TLSRepository(cfg.tls_repository,
                            cfg.openssl_cnf,
                            cfg.server_hostname,
                            cfg.proxy_hostname,
                            clear=create_certificates_for_devices,
                            generate_admin_cert=cfg.generate_admin_cert)

    if create_certificates_for_devices:
        already_represented = set()


        # registers the devices, but doesn't initialize_device the end devices here.
        for k in cfg.devices:
            if tlsrepo.has_device(k.id):
                already_represented.add(k)
            else:
                tlsrepo.create_cert(k.id)
    return tlsrepo


def _shutdown():
    make_stop_file()
    sleep(1)
    remove_stop_file()


def should_stop():
    return Path('server.stop').exists()


def make_stop_file():
    with open('server.stop', 'w') as w:
        pass


def remove_stop_file():
    pth = Path('server.stop')
    if pth.exists():
        os.remove(pth)



def _main():
    parser = ArgumentParser()

    parser.add_argument(dest="config", help="Configuration file for the server.")
    parser.add_argument("--no-validate",
                        action="store_true",
                        help="Allows faster startup since the resolving of addresses is not done!")
    parser.add_argument(
        "--no-create-certs",
        action="store_true",
        help="If specified certificates for for client and server will not be created.")
    parser.add_argument("--debug", action="store_true", help="Debug level of the server")
    parser.add_argument("--production",
                        action="store_true",
                        default=False,
                        help="Run the server in a threaded environment.")
    parser.add_argument("--lfdi",
                        help="Use lfdi mode allows a single lfdi to be connected to on an http connection")
    parser.add_argument("--show-lfdi", action="store_true",
                        help="Show all of the lfdi for the generated certificates and exit.")
    opts = parser.parse_args()

    logging_level = logging.DEBUG if opts.debug else logging.INFO
    logging.basicConfig(level=logging_level)
    logging.getLogger("watchdog.observers.inotify_buffer").setLevel(logging.INFO)

    os.environ["IEEE_2030_5_CONFIG_FILE"] = str(
        Path(opts.config).expanduser().resolve(strict=True))

    cfg_dict = yaml.safe_load(Path(opts.config).expanduser().resolve(strict=True).read_text())

    config = ServerConfiguration(**cfg_dict)

    if config.lfdi_mode == "lfdi_mode_from_file":
        os.environ["IEEE_2030_5_CERT_FROM_COMBINED_FILE"] = '1'

    assert config.tls_repository
    assert config.server_hostname

    add_href(hrefs.get_server_config_href(), config)
    unknown = []
    # Only check for resolvability if not passed --no-validate
    if not opts.no_validate:
        _log.debug("Validating hostnames and/or ip of devices are resolvable.")
        for i in range(len(config.devices)):
            assert config.devices[i].hostname

            try:
                socket.gethostbyname(config.devices[i].hostname)
            except socket.gaierror:
                if hasattr(config.devices[i], "ip"):
                    try:
                        socket.gethostbyname(config.devices[i].ip)
                    except socket.gaierror:
                        unknown.append(config.devices[i].hostname)
                else:
                    unknown.append(config.devices[i].hostname)

    if unknown:
        _log.error("Couldn't resolve the following hostnames.")
        for host in unknown:
            _log.error(host)
        sys.exit(1)

    if opts.show_lfdi and not opts.no_create_certs:
        sys.stderr.write("Can't show lfdi when creating certificates.\n")
        sys.exit(1)

    create_certs = not opts.no_create_certs
    tls_repo = get_tls_repository(config, create_certs)

    if config.gridappsd is not None and create_certs:

        from gridappsd import GridAPPSD

        gapps = GridAPPSD(stomp_address=config.gridappsd.address,
                        stomp_port=config.gridappsd.port,
                        username=config.gridappsd.username,
                        password=config.gridappsd.password)
        assert gapps.connected

        gridappsd_adpt = GridAPPSDAdapter(gapps, config.gridappsd.model_name)
        gridappsd_adpt.create_der_client_certificates(tls_repo)

    if opts.show_lfdi:
        for cn in config.devices:
            sys.stdout.write(f"{cn.id} {tls_repo.lfdi(cn.id)}\n")
        sys.exit(0)

    # Puts the server into http single client lfdi mode.
    if opts.lfdi:
        config.lfdi_client = opts.lfdi

    # Initialize the data storage for the adapters
    if config.storage_path is None:
        config.storage_path = Path("data_store")
    else:
        config.storage_path = Path(config.storage_path)

    # Cleanse means we want to reload the storage each time the server
    # is run.  Note this is dependent on the adapter being filestore
    # not database.  I will have to modify later to deal with that.
    if config.cleanse_storage and config.storage_path.exists():
        _log.debug(f"Removing {config.storage_path}")
        shutil.rmtree(config.storage_path)

    data_store_userdir = Path("~/.ieee_2030_5_data").expanduser()
    if config.cleanse_storage and data_store_userdir.exists():
        _log.debug(f"Removing {data_store_userdir}")
        shutil.rmtree(data_store_userdir)


    # Has to be after we remove the storage path if necessary
    from ieee_2030_5.server.server_constructs import initialize_2030_5

    initialize_2030_5(config, tls_repo)


    from ieee_2030_5.flask_server import run_server

    #from ieee_2030_5.gui import run_gui
    #if not opts.production:
    # #try:
    # p_server = Process(target=run_server,
    #                     kwargs=dict(
    #                         config=config,
    #                         tlsrepo=tls_repo,
    #                         debug=opts.debug,
    #                         use_reloader=False,
    #                         use_debugger=opts.debug,
    #                         threaded=False))
    # p_server.daemon = True
    # p_server.start()
    # if __name__ in {"__main__", "__mp_main__"}:
    #     ui = run_gui()
    #     ui.run_with(app)

    #run_gui()

    # p_gui = Process(target = run_gui)
    # p_gui.daemon = True
    # p_gui.start()


    # # while True:
    # #     sleep(1)
    run_server(config,
                tls_repo,
                debug=opts.debug,
                use_reloader=True,
                use_debugger=opts.debug,
                threaded=False)
    # except KeyboardInterrupt:
    #     _log.info("Shutting down server")
    # finally:
    #     _log.info("Ending Server.")
    # else:
    #     server = build_server(config, tls_repo, enddevices=end_devices)

    #     thread = None
    #     try:
    #         remove_stop_file()
    #         thread = ServerThread(server)
    #         thread.start()
    #         while not should_stop():
    #             sleep(0.5)
    #     except KeyboardInterrupt as ex:
    #         _log.info("Exiting program.")
    #     finally:
    #         if thread:
    #             thread.shutdown()
    #             thread.join()


if __name__ == '__main__':
    try:
        # from werkzeug.serving import is_running_from_reloader
        # print(is_running_from_reloader())
        #if os.environ.get('WERKZEUG_RUN_MAIN') == 'true':
        _main()
    except InvalidConfigFile as ex:
        print(ex.args[0])
    except KeyboardInterrupt:
        pass



from __future__ import annotations

import inspect
import logging
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, TypeVar, Union

import yaml
from dataclasses_json import dataclass_json

__all__ = ["ServerConfiguration", "ReturnValue"]

#from ieee_2030_5.adapters import DERControlAdapter

try:
    from gridappsd.field_interface import MessageBusDefinition
except ImportError as ex:
    pass

import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.server.exceptions import NotFoundError
from ieee_2030_5.types_ import Lfdi

_log = logging.getLogger(__name__)

D = TypeVar("D")


@dataclass
class ReturnValue:
    success: bool
    an_object: D
    was_update: bool
    location: str = None

    def get(self, datatype: D) -> D:
        return self.an_object


class InvalidConfigFile(Exception):
    pass


@dataclass
class FSAConfiguration:
    description: str
    programs: List[ProgramConfiguration] = field(default_factory=list)


@dataclass
class DERConfiguration:
    #capabilities:
    modesSupported: str
    type: int


@dataclass
class DeviceConfiguration:
    id: str = None
    post_rate: int = 900
    pin: int = None
    poll_rate: int = 900
    fsas: List[str] = field(default_factory=list)
    ders: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.id.__hash__()


@dataclass
class CurveConfiguration:

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class ControlBaseConfiguration:

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class ControlConfiguration:
    description: str = None
    base: Dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, env):
        return cls(
            **{
                k: v
                for k, v in env.items() if k in inspect.signature(m.DERControl).parameters
                or k in inspect.signature(cls).parameters
            })

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class ProgramConfiguration:

    default_control: str = None
    controls: List[str] = field(default_factory=list)
    curves: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass_json
@dataclass
class GridappsdConfiguration:
    model_name: str
    address: str = 'localhost'
    port: int = 61613
    username: str = 'system'
    password: str = 'manager'
    field_bus_config: Optional[str] = None
    field_bus_def: Optional[MessageBusDefinition] = None
    feeder_id_file: Optional[str] = None
    feeder_id: Optional[str] = None
    simulation_id_file: Optional[str] = None
    simulation_id: Optional[str] = None


# @dataclass
# class ControlDefaults:
#     setESDelay: Optional[Int]
#     setESHighFreq: UInt16 [0..1]
#     setESHighVolt: Int16 [0..1]
#     setESLowFreq: UInt16 [0..1]
#     setESLowVolt: Int16 [0..1]
#     setESRampTms: UInt32 [0..1]
#     setESRandomDelay: UInt32 [0..1]
#     setGradW: UInt16 [0..1]
#     setSoftGradW: UInt16 [0..1]

# @dataclass
# class Program:
#     description: str
#     mRID: Optional[str]

# @dataclass
# class Control:
#     description: str
#     mRID: Optional[str]


@dataclass
class ProgramList:
    name: str
    programs: List[m.DERProgram]


@dataclass
class ServerConfiguration:
    openssl_cnf: str

    tls_repository: str
    openssl_cnf: str

    server: str
    port: int
    ui_port: int = None

    cleanse_storage: bool = True
    storage_path: str = None

    log_event_list_poll_rate: int = 900
    device_capability_poll_rate: int = 900
    mirror_usage_point_post_rate: int = 300
    end_device_list_poll_rate: int = 86400    # daily check-in

    generate_admin_cert: bool = False
    lfdi_client: str = None

    fsas: List[FSAConfiguration] = field(default_factory=list)
    programs: List[ProgramConfiguration] = field(default_factory=list)
    devices: List[DeviceConfiguration] = field(default_factory=list)
    ders: List[DERConfiguration] = field(default_factory=list)
    curves: List[CurveConfiguration] = field(default_factory=list)

    server_mode: Union[
        Literal["enddevices_create_on_start"],
        Literal["enddevices_register_access_only"]] = "enddevices_register_access_only"

    lfdi_mode: Union[
        Literal["lfdi_mode_from_file"],
        Literal["lfdi_mode_from_cert_fingerprint"]] = "lfdi_mode_from_cert_fingerprint"

    # programs: List[DERProgramConfiguration] = field(default_factory=list)
    # controls: List[DERControlConfiguration] = field(default_factory=list)
    # curves: List[DERCurveConfiguration] = field(default_factory=list)
    # events: List[Dict] = field(default_factory=list)

    # # map into program_lists array for programs for specific
    # # named list.
    # programs_map: Dict[str, int] = field(default_factory=dict)
    # program_lists: List[ProgramList] = field(default_factory=list)
    # fsa_list: List[FunctionSetAssignments] = field(default_factory=list)
    # curve_list: List[DERCurve] = field(default_factory=list)

    proxy_hostname: Optional[str] = None
    gridappsd: Optional[GridappsdConfiguration] = None
    # DefaultDERControl: Optional[DefaultDERControl] = None
    # DERControlList: Optional[DERControl] = field(default=list)

    @property
    def server_hostname(self) -> str:
        server = self.server
        if self.port:
            server = server + f":{self.port}"

        return server

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def load(cls, file: Path) -> ServerConfiguration:
        if not file.exists():
            raise InvalidConfigFile(f"File does not exist: {file}")
        return cls.from_dict(yaml.safe_load(file.read_text()))

    def __post_init__(self):
        # self.curves = [DERCurveConfiguration.from_dict(x) for x in self.curves]
        # self.controls = [DERControlConfiguration.from_dict(x) for x in self.controls]
        # self.programs = [DERProgramConfiguration.from_dict(x) for x in self.programs]

        if self.devices is None:
            self.devices = []
        else:
            self.devices = [DeviceConfiguration.from_dict(x) for x in self.devices]
        # for d in self.devices:
        # d.deviceCategory = eval(f"m.DeviceCategoryType.{d.deviceCategory}").name
        #d.device_category_type = eval(f"m.DeviceCategoryType.{d.device_category_type}")

        # der_controls, der_default_control = None, None
        # if self.DERControlListFile:
        #     der_controls, der_default_control = DERControlAdapter.load_from_yaml_file(self.DERControlListFile)

        # temp_program_list = self.program_lists
        # if isinstance(self.program_lists, str):
        #     temp_program_list = yaml.safe_load(Path(self.program_lists).read_text())

        # self.program_lists = []
        # for program_list in temp_program_list['program_lists']:
        #     pl_obj = ProgramList(name=program_list["name"], programs=[])
        #     for inter_index, program_obj in enumerate(program_list['programs']):
        #         base = None
        #         if "DERControlBase" in program_obj:
        #             base = DERControlBase()
        #             for k in program_obj.get("DERControlBase"):
        #                 setattr(base, k, program_obj["DERControlBase"].get(k))
        #             del program_obj["DERControlBase"]
        #         # TODO Do Something with base so we can retrieve it.
        #         # if base:
        #         #     self.DefaultDERControl = DefaultDERControl(DERControlBase=base)
        #         #     for k, v in program.items():
        #         #         setattr(self.DefaultDERControl, k, v)
        #         # else:
        #         program = DERProgram()
        #
        #         for k, v in program_obj.items():
        #             setattr(program, k, v)
        #
        #         pl_obj.programs.append(program)
        #     self.program_lists.append(pl_obj)
        #
        # temp_curve_list = self.curve_list
        # if isinstance(self.curve_list, str):
        #     temp_curve_list = yaml.safe_load(Path(self.curve_list).read_text())
        #
        # self.curve_list = []
        # for item in temp_curve_list['curve_list']:
        #     curve_data: List[CurveData] = []
        #     for data in item.get('CurveData'):
        #         curve_data.append(CurveData(xvalue=data['xvalue'], yvalue=data['yvalue']))
        #     del item["CurveData"]
        #
        #     curve = DERCurve(CurveData=curve_data)
        #     for k, v in item.items():
        #         setattr(curve, k, v)
        #     self.curve_list.append(curve)

        if self.gridappsd:
            self.gridappsd = GridappsdConfiguration.from_dict(self.gridappsd)

            # TODO Configuration for field bus here
            # if Path(self.gridappsd.feeder_id_file).exists():
            #     self.gridappsd.feeder_id = Path(self.gridappsd.feeder_id_file).read_text().strip()
            # if Path(self.gridappsd.simulation_id_file).exists():
            #     self.gridappsd.simulation_id = Path(
            #         self.gridappsd.simulation_id_file).read_text().strip()

            # if not self.gridappsd.feeder_id:
            #     raise ValueError(
            #         "Feeder id from gridappsd not found in feeder_id_file nor was specified "
            #         "in gridappsd config section.")

            # # TODO: This might not be the best place for this manipulation
            # self.gridappsd.field_bus_def = MessageBusDefinition.load(
            #     self.gridappsd.field_bus_config)
            # self.gridappsd.field_bus_def.id = self.gridappsd.feeder_id

            # _log.info("Gridappsd Configuration For Simulation")
            # _log.info(f"feeder id: {self.gridappsd.feeder_id}")
            # if self.gridappsd.simulation_id:
            #     _log.info(f"simulation id: {self.gridappsd.simulation_id}")
            # else:
            #     _log.info("no simulation id")
            # _log.info("x" * 80)

        # if self.field_bus_config:
        #     self.field_bus_def = MessageBusDefinition.load(self.field_bus_config)

    def get_device_pin(self, lfdi: Lfdi, tls_repo: TLSRepository) -> int:
        for d in self.devices:
            test_lfdi = tls_repo.lfdi(d.id)
            if test_lfdi == int(lfdi):
                return d.pin
        raise NotFoundError(f"The device_id: {lfdi} was not found.")
